from typing import List, Optional

import attrs
from web_poet import Returns

from zyte_common_items.components import Breadcrumb
from zyte_common_items.fields import auto_field
from zyte_common_items.items import ArticleFromList, ArticleList, ArticleListMetadata
from zyte_common_items.processors import breadcrumbs_processor

from .base import BasePage, Page
from .mixins import HasMetadata


class BaseArticleListPage(
    BasePage, Returns[ArticleList], HasMetadata[ArticleListMetadata]
):
    """:class:`BasePage` subclass for :class:`ArticleList`."""

    class Processors(BasePage.Processors):
        breadcrumbs = [breadcrumbs_processor]


class ArticleListPage(Page, Returns[ArticleList], HasMetadata[ArticleListMetadata]):
    """:class:`Page` subclass for :class:`ArticleList`."""

    class Processors(Page.Processors):
        breadcrumbs = [breadcrumbs_processor]


@attrs.define
class AutoArticleListPage(BaseArticleListPage):
    article_list: ArticleList

    @auto_field
    def articles(self) -> Optional[List[ArticleFromList]]:
        return self.article_list.articles

    @auto_field
    def breadcrumbs(self) -> Optional[List[Breadcrumb]]:
        return self.article_list.breadcrumbs

    @auto_field
    def canonicalUrl(self) -> Optional[str]:
        return self.article_list.canonicalUrl

    @auto_field
    def metadata(self) -> Optional[ArticleListMetadata]:
        return self.article_list.metadata

    @auto_field
    def url(self) -> Optional[str]:
        return self.article_list.url



from typing import Optional

import attrs

from zyte_common_items.base import Item


@attrs.define(kw_only=True)
class AggregateRating(Item):
    """Aggregate data about reviews and ratings.

    At least one of :attr:`ratingValue` or :attr:`reviewCount` is required.

    See :attr:`Product.aggregateRating
    <zyte_common_items.Product.aggregateRating>`.
    """

    #: Maximum value of the rating system.
    bestRating: Optional[float] = None

    #: Average value of all ratings.
    ratingValue: Optional[float] = None

    #: Review count.
    reviewCount: Optional[int] = None


@attrs.define(kw_only=True)
class StarRating(Item):
    """Official star rating of a place."""

    #: Star rating of the place, as it appears on the page, without processing.
    raw: Optional[str] = None

    #: Star rating value of the place.
    ratingValue: Optional[float] = None



import pytest
from web_poet import HttpResponse, field

from zyte_common_items import ProductPage


@pytest.mark.asyncio
async def test_price_selector():
    class CustomProductPage(ProductPage):
        call_count = 0

        @field
        def price(self):
            self.call_count += 1
            return self.css("div::text")

    html = b"""
    <!DOCTYPE html>
    <html>
        <body>
            <div>$13.2</div>
        </body>
    </html>
    """
    url = "https://example.com"
    page = CustomProductPage(response=HttpResponse(url=url, body=html))
    assert page.call_count == 0
    assert page.price == "13.20"
    assert page.call_count == 1
    assert page.currency is None
    assert await page.currencyRaw == "$"
    assert page.call_count == 1

    # access currency fields before the price field
    page = CustomProductPage(response=HttpResponse(url=url, body=html))
    assert page.call_count == 0
    assert page.currency is None
    assert await page.currencyRaw == "$"
    assert page.call_count == 1
    assert page.price == "13.20"
    assert page.call_count == 2  # we want this to be 1


@pytest.mark.asyncio
async def test_price_explicit():
    class CustomProductPage(ProductPage):
        call_count = 0

        @field
        def price(self):
            self.call_count += 1
            return self.css("div::text").get()

    html = b"""
    <!DOCTYPE html>
    <html>
        <body>
            <div>$13.2</div>
        </body>
    </html>
    """
    url = "https://example.com"
    page = CustomProductPage(response=HttpResponse(url=url, body=html))
    assert page.call_count == 0
    assert page.price == "$13.2"
    assert page.call_count == 1
    assert page.currency is None
    assert await page.currencyRaw is None
    assert page.call_count == 2  # we want this to be 1
    assert await page.currencyRaw is None
    assert page.call_count == 2  # we want this to be 1

    # access currency fields before the price field
    page = CustomProductPage(response=HttpResponse(url=url, body=html))
    assert page.call_count == 0
    assert page.currency is None
    assert await page.currencyRaw is None
    assert page.call_count == 1
    assert page.price == "$13.2"
    assert page.call_count == 2  # we want this to be 1


@pytest.mark.asyncio
async def test_price_async():
    class CustomProductPage(ProductPage):
        @field
        async def price(self):
            return self.css("div::text")

    html = b"""
    <!DOCTYPE html>
    <html>
        <body>
            <div>$13.2</div>
        </body>
    </html>
    """
    url = "https://example.com"
    page = CustomProductPage(response=HttpResponse(url=url, body=html))
    assert await page.price == "13.20"
    assert page.currency is None
    assert await page.currencyRaw == "$"

    # access currency fields before the price field
    page = CustomProductPage(response=HttpResponse(url=url, body=html))
    assert page.currency is None
    assert await page.currencyRaw == "$"
    assert await page.price == "13.20"


@pytest.mark.asyncio
async def test_currency_hardcoded():
    class CustomProductPage(ProductPage):
        CURRENCY = "USD"

        @field
        def price(self):
            return self.css("div::text")

    html = b"""
    <!DOCTYPE html>
    <html>
        <body>
            <div>$13.2</div>
        </body>
    </html>
    """
    url = "https://example.com"
    page = CustomProductPage(response=HttpResponse(url=url, body=html))
    assert page.price == "13.20"
    assert page.currency == "USD"
    assert await page.currencyRaw == "$"

    # access currency fields before the price field
    page = CustomProductPage(response=HttpResponse(url=url, body=html))
    assert page.currency == "USD"
    assert await page.currencyRaw == "$"
    assert page.price == "13.20"


@pytest.mark.asyncio
async def test_currency_no_price():
    class CustomProductPage(ProductPage):
        CURRENCY = "USD"

    html = b"""
    <!DOCTYPE html>
    <html>
        <body>
            <div>$13.2</div>
        </body>
    </html>
    """
    url = "https://example.com"
    page = CustomProductPage(response=HttpResponse(url=url, body=html))
    assert page.currency == "USD"
    assert await page.currencyRaw is None


@pytest.mark.asyncio
async def test_currency_none_price():
    class CustomProductPage(ProductPage):
        CURRENCY = "USD"
        call_count = 0

        @field
        def price(self):
            self.call_count += 1
            return None

    html = b"""
    <!DOCTYPE html>
    <html>
        <body>
            <div>$13.2</div>
        </body>
    </html>
    """
    url = "https://example.com"
    page = CustomProductPage(response=HttpResponse(url=url, body=html))
    assert page.call_count == 0
    assert page.currency == "USD"
    assert await page.currencyRaw is None
    assert page.call_count == 1
    assert await page.currencyRaw is None
    assert page.call_count == 1
    assert page.price is None
    assert page.call_count == 2  # we want this to be 1


@pytest.mark.asyncio
async def test_regularPrice():
    class CustomProductPage(ProductPage):
        CURRENCY = "USD"

        @field
        def price(self):
            return self.css("b::text")

        @field
        def regularPrice(self):
            return self.css("div::text")

    html = b"""
    <!DOCTYPE html>
    <html>
        <body>
            <div>$13.2</div>
            <b>$10.2</b>
        </body>
    </html>
    """
    url = "https://example.com"
    page = CustomProductPage(response=HttpResponse(url=url, body=html))
    assert page.regularPrice == "13.20"
    assert page.price == "10.20"
    assert page.currency == "USD"
    assert await page.currencyRaw == "$"

    # access currency fields before the price field
    page = CustomProductPage(response=HttpResponse(url=url, body=html))
    assert page.currency == "USD"
    assert await page.currencyRaw == "$"
    assert page.price == "10.20"
    assert page.regularPrice == "13.20"


@pytest.mark.asyncio
async def test_price_invalid():
    class CustomProductPage(ProductPage):
        CURRENCY = "USD"

        @field
        def price(self):
            return self.css("div::text")

    html = b"""
    <!DOCTYPE html>
    <html>
        <body>
            <div>Out of stock</div>
        </body>
    </html>
    """
    url = "https://example.com"
    page = CustomProductPage(response=HttpResponse(url=url, body=html))
    assert page.price is None
    assert page.currency == "USD"
    assert await page.currencyRaw is None

    # access currency fields before the price field
    page = CustomProductPage(response=HttpResponse(url=url, body=html))
    assert page.currency == "USD"
    assert await page.currencyRaw is None
    assert page.price is None


def test_currencyRaw_explicit():
    class CustomProductPage(ProductPage):
        @field
        def price(self):
            return self.css("div::text")

        @field
        def currencyRaw(self):
            return "US$"

    html = b"""
    <!DOCTYPE html>
    <html>
        <body>
            <div>$13.2</div>
        </body>
    </html>
    """
    url = "https://example.com"
    page = CustomProductPage(response=HttpResponse(url=url, body=html))
    assert page.price == "13.20"
    assert page.currency is None
    assert page.currencyRaw == "US$"

    # access currency fields before the price field
    page = CustomProductPage(response=HttpResponse(url=url, body=html))
    assert page.currency is None
    assert page.currencyRaw == "US$"
    assert page.price == "13.20"



import datetime
import sys

try:
    from datetime import UTC
except ImportError:
    # Python < 3.11
    from datetime import timezone

    UTC = timezone.utc


def format_datetime(dt) -> str:
    """Return the specified :class:`~datetime.datetime` object, assumed to be
    in the UTC timezone, in ISO format, with the timezone specified as
    ``Z``."""
    return f"{dt.replace(tzinfo=None).isoformat(timespec='seconds')}Z"


def parse_iso_datetime(date_str) -> datetime.datetime:
    """
    Parse ISO-formatted UTC date (with a timezone specified as Z)
    to a TZ-aware datetime object.
    """
    if sys.version_info < (3, 11):
        assert date_str[-1] == "Z"
        return datetime.datetime.fromisoformat(date_str[:-1]).replace(tzinfo=UTC)
    return datetime.datetime.fromisoformat(date_str)


def utcnow_formatted() -> str:
    return format_datetime(utcnow())


def utcnow() -> datetime.datetime:
    """Return current datetime in UTC, as a tz-aware datetime object"""
    return datetime.datetime.now(UTC)



"""
A module with common attrs converters
"""

from typing import Optional, Union

import attrs
from web_poet.page_inputs.url import _Url


def url_to_str(url: Union[str, _Url]) -> str:
    """Return the input :class:`~web_poet.page_inputs.http.RequestUrl` or
    :class:`~web_poet.page_inputs.http.ResponseUrl` object as a string."""

    if not isinstance(url, (str, _Url)):
        raise ValueError(
            f"{url!r} is neither a string nor an instance of RequestUrl or ResponseUrl."
        )
    return str(url)


def url_to_str_optional(url: Union[str, _Url, None]) -> Optional[str]:
    """Return the input :class:`~web_poet.page_inputs.http.RequestUrl` or
    :class:`~web_poet.page_inputs.http.ResponseUrl` object as a string, or
    None if url is None."""
    if url is None:
        return None
    return url_to_str(url)


class MetadataCaster:
    """attrs converter that converts an input metadata object into the metadata
    class declared by the container page object class."""

    def __init__(self, target):
        self._target = target

    def __call__(self, value):
        return value.cast(self._target)


def to_probability_request_list(request_list):
    """attrs converter to turn lists of :class:`~scrapy.http.Request` instances into
    lists of :class:`~.ProbabilityRequest` instances."""
    from zyte_common_items.components import ProbabilityRequest

    return [request.cast(ProbabilityRequest) for request in request_list]


def to_probability_request_list_optional(request_list):
    """attrs converter to turn lists of :class:`~scrapy.http.Request` instances into
    lists of :class:`~.ProbabilityRequest` instances. If None is passed, None is
    returned."""
    if request_list is None:
        return None
    return to_probability_request_list(request_list)


def to_metadata(metadata_cls: type):
    return MetadataCaster(metadata_cls)


def to_metadata_optional(metadata_cls: type):
    return attrs.converters.optional(MetadataCaster(metadata_cls))



from typing import Optional

import attrs

from zyte_common_items.base import Item
from zyte_common_items.util import url_to_str


@attrs.define(kw_only=True)
class Author(Item):
    """Author of an article.

    See :attr:`Article.authors <zyte_common_items.Article.authors>`.
    """

    #: Email.
    email: Optional[str] = None

    #: URL of the details page of the author.
    url: Optional[str] = attrs.field(
        default=None, converter=attrs.converters.optional(url_to_str), kw_only=True
    )

    #: Full name.
    name: Optional[str] = None

    #: Text from which :attr:`~zyte_common_items.Author.name` was
    #: extracted.
    nameRaw: Optional[str] = None



import attrs

from zyte_common_items.base import Item


@attrs.define
class AdditionalProperty(Item):
    """A name-value pair.

    See :attr:`Product.additionalProperties
    <zyte_common_items.Product.additionalProperties>`.
    """

    #: Name.
    name: str

    #: Value.
    value: str



from typing import List, Optional

import attrs

from zyte_common_items.base import Item
from zyte_common_items.components import Metadata, Reactions, SocialMediaPostAuthor, Url
from zyte_common_items.converters import to_metadata_optional, url_to_str


@attrs.define(kw_only=True)
class SocialMediaPostMetadata(Metadata):
    """Metadata class for :data:`zyte_common_items.SocialMediaPost.metadata`."""


@attrs.define(kw_only=True)
class SocialMediaPost(Item):
    """Represents a single social media post."""

    #: The URL of the final response, after any redirects.
    url: str = attrs.field(converter=url_to_str)

    #: The identifier of the post.
    postId: Optional[str] = None

    #: Details of reactions to the post.
    reactions: Optional[Reactions] = None

    #: The text content of the post.
    text: Optional[str] = None

    #: The timestamp at which the post was created.
    #:
    #: Format: Timezone: UTC. ISO 8601 format: "YYYY-MM-DDThh:mm:ssZ"
    datePublished: Optional[str] = None

    #: The list of hashtags contained in the post.
    hashtags: Optional[List[str]] = None

    #: The list of URLs of media files (images, videos, etc.) linked from the post.
    mediaUrls: Optional[List[Url]] = None

    #: Details of the author of the post.
    #:
    #: No easily identifiable information can be contained in here, such as usernames.
    author: Optional[SocialMediaPostAuthor] = None

    #: Contains metadata about the data extraction process.
    metadata: Optional[SocialMediaPostMetadata] = attrs.field(
        default=None, converter=to_metadata_optional(SocialMediaPostMetadata), kw_only=True  # type: ignore[misc]
    )



from typing import List, Optional

import attrs
from web_poet import Returns

from zyte_common_items.components import Audio, Author, Breadcrumb, Image, Video
from zyte_common_items.fields import auto_field
from zyte_common_items.items import Article, ArticleMetadata
from zyte_common_items.processors import breadcrumbs_processor, images_processor

from .base import BasePage, Page
from .mixins import HasMetadata


class BaseArticlePage(BasePage, Returns[Article], HasMetadata[ArticleMetadata]):
    """:class:`BasePage` subclass for :class:`Article`."""

    class Processors(BasePage.Processors):
        breadcrumbs = [breadcrumbs_processor]
        images = [images_processor]


class ArticlePage(Page, Returns[Article], HasMetadata[ArticleMetadata]):
    """:class:`Page` subclass for :class:`Article`."""

    class Processors(Page.Processors):
        breadcrumbs = [breadcrumbs_processor]
        images = [images_processor]


@attrs.define
class AutoArticlePage(BaseArticlePage):
    article: Article

    @auto_field
    def headline(self) -> Optional[str]:
        return self.article.headline

    @auto_field
    def datePublished(self) -> Optional[str]:
        return self.article.datePublished

    @auto_field
    def datePublishedRaw(self) -> Optional[str]:
        return self.article.datePublishedRaw

    @auto_field
    def dateModified(self) -> Optional[str]:
        return self.article.dateModified

    @auto_field
    def dateModifiedRaw(self) -> Optional[str]:
        return self.article.dateModifiedRaw

    @auto_field
    def authors(self) -> Optional[List[Author]]:
        return self.article.authors

    @auto_field
    def breadcrumbs(self) -> Optional[List[Breadcrumb]]:
        return self.article.breadcrumbs

    @auto_field
    def inLanguage(self) -> Optional[str]:
        return self.article.inLanguage

    @auto_field
    def mainImage(self) -> Optional[Image]:
        return self.article.mainImage

    @auto_field
    def images(self) -> Optional[List[Image]]:
        return self.article.images

    @auto_field
    def description(self) -> Optional[str]:
        return self.article.description

    @auto_field
    def articleBody(self) -> Optional[str]:
        return self.article.articleBody

    @auto_field
    def articleBodyHtml(self) -> Optional[str]:
        return self.article.articleBodyHtml

    @auto_field
    def videos(self) -> Optional[List[Video]]:
        return self.article.videos

    @auto_field
    def audios(self) -> Optional[List[Audio]]:
        return self.article.audios

    @auto_field
    def canonicalUrl(self) -> Optional[str]:
        return self.article.canonicalUrl

    @auto_field
    def url(self) -> Optional[str]:
        return self.article.url

    @auto_field
    def metadata(self) -> Optional[ArticleMetadata]:
        return self.article.metadata



import datetime

from zyte_common_items._dateutils import format_datetime, parse_iso_datetime


def test_parse_and_format():
    dt_str = "2024-04-11T15:06:02Z"
    dt_obj = datetime.datetime(
        year=2024,
        month=4,
        day=11,
        hour=15,
        minute=6,
        second=2,
        microsecond=0,
        tzinfo=datetime.timezone.utc,
    )

    assert parse_iso_datetime(dt_str) == dt_obj
    assert format_datetime(dt_obj) == dt_str



from __future__ import annotations

from typing import Optional

import attrs

from zyte_common_items.base import Item


@attrs.define(kw_only=True)
class JobLocation(Item):
    """Location of a job offer."""

    #: Job location, as it appears on the website.
    raw: Optional[str] = None


@attrs.define(kw_only=True)
class BaseSalary(Item):
    """Base salary of a job offer."""

    #: Salary amount as it appears on the website.
    raw: Optional[str] = None

    #: The minimum value of the base salary as a number string.
    valueMin: Optional[str] = None

    #: The maximum value of the base salary as a number string.
    valueMax: Optional[str] = None

    #: The type of rate associated with the salary, e.g. monthly, annual, daily.
    rateType: Optional[str] = None

    #: Currency associated with the salary amount.
    currency: Optional[str] = None

    #: Currency associated with the salary amount, without normalization.
    currencyRaw: Optional[str] = None


@attrs.define(kw_only=True)
class HiringOrganization(Item):
    """Organization that is hiring for a job offer."""

    #: Name of the hiring organization.
    name: Optional[str] = None

    #: Organization information as available on the website.
    nameRaw: Optional[str] = None

    #: Identifier of the organization used by job posting website.
    id: Optional[str] = None



from collections.abc import Collection
from contextlib import contextmanager
from copy import copy

# In Python ≤ 3.8 you cannot annotate with “collections.abc.Collection[Item]”,
# so we need to import typing.Collection for annotation instead.
from typing import Collection as CollectionType
from typing import Optional

import attrs
import pytest
from itemadapter import ItemAdapter

from zyte_common_items import Item, Product, ZyteItemAdapter
from zyte_common_items.adapter import ZyteItemKeepEmptyAdapter

from .test_items import _PRODUCT_ALL_KWARGS, _PRODUCT_MIN_KWARGS


@contextmanager
def configured_adapter(adapter=ZyteItemAdapter):
    original_value = copy(ItemAdapter.ADAPTER_CLASSES)
    ItemAdapter.ADAPTER_CLASSES = (adapter, *ItemAdapter.ADAPTER_CLASSES)
    try:
        yield
    finally:
        ItemAdapter.ADAPTER_CLASSES = original_value


def test_asdict_all_fields():
    product = Product(**_PRODUCT_ALL_KWARGS)
    with configured_adapter():
        adapter = ItemAdapter(product)
        actual_dict = adapter.asdict()
    expected_dict = {
        "additionalProperties": [dict(name="foo", value="bar")],
        "aggregateRating": dict(
            bestRating=5.0,
            ratingValue=2.5,
            reviewCount=123,
        ),
        "availability": "InStock",
        "brand": dict(name="Ka-pow"),
        "breadcrumbs": [
            dict(name="Level 1", url="http://example.com/level1"),
            dict(name="Level 2", url="http://example.com/level1/level2"),
        ],
        "canonicalUrl": "https://example.com/product22",
        "color": "white",
        "currency": "USD",
        "currencyRaw": "$",
        "description": "Full freshness all over the fridge\n5 Conversion Modes on demand\nSuper Cooling Plus™",
        "descriptionHtml": (
            "<article><p>Full freshness all over the fridge</p>"
            "<p>5 Conversion Modes on demand</p>"
            "<p>Super Cooling Plus&trade;</p></article>"
        ),
        "features": ["Easily store fragile products.", "Bluetooth connectivity."],
        "gtin": [dict(type="foo", value="bar")],
        "images": [
            dict(url="http://example.com/image1.png"),
        ],
        "mainImage": dict(url="http://example.com/image1.png"),
        "metadata": dict(
            dateDownloaded="2022-12-31T13:01:54Z",
            probability=1.0,
        ),
        "mpn": "HSC0424PP",
        "name": "White two-door refrigerator",
        "price": "9999.99",
        "productId": "A123DK9823",
        "regularPrice": "11999.99",
        "size": "XL",
        "sku": "A123DK9823",
        "style": "polka dots",
        "url": "https://example.com/?product=product22",
        "variants": [{}],
    }
    assert actual_dict == expected_dict


def test_product_asdict_min():
    product = Product(**_PRODUCT_MIN_KWARGS)
    with configured_adapter():
        adapter = ItemAdapter(product)
        actual_dict = adapter.asdict()
    expected_dict = {
        "url": "https://example.com/?product=product22",
    }
    assert actual_dict == expected_dict


def test_product_asdict_list():
    product = Product(**_PRODUCT_MIN_KWARGS, variants=[{}])  # type: ignore[list-item]
    with configured_adapter():
        adapter = ItemAdapter(product)
        actual_dict = adapter.asdict()
    expected_dict = {
        "url": "https://example.com/?product=product22",
        "variants": [{}],
    }
    assert actual_dict == expected_dict


class EmptyCollection(Collection):
    def __contains__(self, item):
        return False

    def __len__(self):
        return 0

    def __iter__(self):
        return
        yield


@pytest.mark.parametrize(
    "value",
    (
        [],
        tuple(),
        EmptyCollection(),
    ),
)
def test_asdict_empty_collection(value):
    @attrs.define
    class _Item(Item):
        children: CollectionType[Item]

    item = _Item(value)
    with configured_adapter():
        adapter = ItemAdapter(item)
        actual_dict = adapter.asdict()
    assert actual_dict == {}


def test_asdict_unknown_fields():
    input_dict = dict(
        a="b",
        additionalProperties=[{"name": "a", "value": "b", "max": 10}],
        aggregateRating={"worstRating": 0},
        url="https://example.com/",
    )
    product = Product.from_dict(input_dict)
    with configured_adapter():
        adapter = ItemAdapter(product)
        assert adapter.asdict() == input_dict


def test_field_meta():
    metadata = {"b": "c"}

    @attrs.define
    class _Item(Item):
        a = attrs.field(default=None, metadata=metadata)

    item = _Item()
    with configured_adapter():
        adapter = ItemAdapter(item)
        actual_dict = adapter.get_field_meta("a")
    assert actual_dict == metadata


def test_field_meta_missing_field():
    item = Item()
    with configured_adapter():
        adapter = ItemAdapter(item)
        with pytest.raises(KeyError):
            adapter.get_field_meta("a")


def test_field_meta_unknown_field():
    """There cannot be metadata for unknown fields, but trying to get it should
    not cause an exception either."""
    item = Item.from_dict({"a": None})
    with configured_adapter():
        adapter = ItemAdapter(item)
        actual_dict = adapter.get_field_meta("a")
    assert actual_dict == {}


def test_field_names():
    @attrs.define
    class _Item(Item):
        a: int
        b: Optional[int] = None

    item = _Item.from_dict({"a": 1, "c": 2, "d": None})
    with configured_adapter():
        adapter = ItemAdapter(item)
        actual = adapter.field_names()
    assert tuple(actual) == ("a", "b", "c", "d")


def test_known_field_get():
    url = "https://example.com/"
    product = Product(url=url)
    with configured_adapter():
        adapter = ItemAdapter(product)
        assert adapter["url"] == url


def test_known_field_get_missing():
    url = "https://example.com/"
    product = Product(url=url)
    with configured_adapter():
        adapter = ItemAdapter(product)
        assert adapter["canonicalUrl"] is None


def test_known_field_set():
    url = "https://example.com/"
    product = Product(url=url)
    with configured_adapter():
        adapter = ItemAdapter(product)
        adapter["canonicalUrl"] = url
        assert adapter["canonicalUrl"] == url
        assert product.canonicalUrl == url


def test_known_field_update():
    product = Product(url="https://example.com/a")
    with configured_adapter():
        adapter = ItemAdapter(product)
        new_url = "https://example.com/b"
        adapter["url"] = new_url
        assert adapter["url"] == new_url
        assert product.url == new_url


def test_known_field_remove():
    url = "https://example.com/"
    product = Product(url=url, canonicalUrl=url)
    with configured_adapter():
        adapter = ItemAdapter(product)
        del adapter["canonicalUrl"]
        assert "canonicalUrl" not in adapter
        assert not hasattr(product, "canonicalUrl")


def test_known_field_remove_missing():
    product = Product(url="https://example.com/")
    with configured_adapter():
        adapter = ItemAdapter(product)
        del adapter["canonicalUrl"]
        assert "canonicalUrl" not in adapter
        assert not hasattr(product, "canonicalUrl")


def test_known_field_remove_missing_twice():
    product = Product(url="https://example.com/")
    with configured_adapter():
        adapter = ItemAdapter(product)
        del adapter["canonicalUrl"]
        with pytest.raises(KeyError):
            del adapter["canonicalUrl"]


def test_unknown_field_get():
    product = Product.from_dict(
        dict(
            a="b",
            additionalProperties=[{"name": "a", "value": "b", "max": 10}],
            aggregateRating={"worstRating": 0},
            url="https://example.com/",
        )
    )

    with configured_adapter():
        adapter = ItemAdapter(product)
        assert adapter["a"] == "b"

        adapter = ItemAdapter(product.additionalProperties[0])
        assert adapter["max"] == 10

        adapter = ItemAdapter(product.aggregateRating)
        assert adapter["worstRating"] == 0


def test_unknown_field_get_missing():
    product = Product(url="https://example.com")
    with configured_adapter():
        adapter = ItemAdapter(product)
        with pytest.raises(KeyError):
            adapter["a"]


def test_unknown_field_set():
    product = Product.from_dict(
        dict(
            additionalProperties=[{"name": "a", "value": "b"}],
            aggregateRating={"bestRating": 5.0},
            url="https://example.com/",
        )
    )

    with configured_adapter():
        adapter = ItemAdapter(product)
        adapter["a"] = "b"
        assert adapter["a"] == "b"
        assert product._unknown_fields_dict["a"] == "b"

        adapter = ItemAdapter(product.additionalProperties[0])
        adapter["max"] = 10
        assert adapter["max"] == 10
        assert product.additionalProperties[0]._unknown_fields_dict["max"] == 10

        adapter = ItemAdapter(product.aggregateRating)
        adapter["worstRating"] = 0
        assert adapter["worstRating"] == 0
        assert product.aggregateRating._unknown_fields_dict["worstRating"] == 0


def test_unknown_field_update():
    product = Product.from_dict(
        dict(
            a="b",
            additionalProperties=[{"name": "a", "value": "b", "max": 10}],
            aggregateRating={"worstRating": 0},
            url="https://example.com/",
        )
    )

    with configured_adapter():
        adapter = ItemAdapter(product)
        adapter["a"] = "c"
        assert adapter["a"] == "c"
        assert product._unknown_fields_dict["a"] == "c"

        adapter = ItemAdapter(product.additionalProperties[0])
        adapter["max"] = 20
        assert adapter["max"] == 20
        assert product.additionalProperties[0]._unknown_fields_dict["max"] == 20

        adapter = ItemAdapter(product.aggregateRating)
        adapter["worstRating"] = 1
        assert adapter["worstRating"] == 1
        assert product.aggregateRating._unknown_fields_dict["worstRating"] == 1


def test_unknown_field_remove():
    product = Product.from_dict(
        dict(
            a="b",
            additionalProperties=[{"name": "a", "value": "b", "max": 10}],
            aggregateRating={"worstRating": 0},
            url="https://example.com/",
        )
    )

    with configured_adapter():
        adapter = ItemAdapter(product)
        del adapter["a"]
        assert "a" not in adapter
        assert "a" not in product._unknown_fields_dict

        adapter = ItemAdapter(product.additionalProperties[0])
        del adapter["max"]
        assert "max" not in adapter
        assert "max" not in product.additionalProperties[0]._unknown_fields_dict

        adapter = ItemAdapter(product.aggregateRating)
        del adapter["worstRating"]
        assert "worstRating" not in adapter
        assert "worstRating" not in product.aggregateRating._unknown_fields_dict


def test_unknown_field_remove_missing():
    product = Product(url="https://example.com/")
    with configured_adapter():
        adapter = ItemAdapter(product)
        with pytest.raises(KeyError):
            del adapter["a"]


def test_keep_empty_adapter_global():
    @attrs.define
    class _Item(Item):
        children: CollectionType[Item]

    item = _Item([])
    with configured_adapter(ZyteItemKeepEmptyAdapter):
        adapter = ItemAdapter(item)
        actual_dict = adapter.asdict()
    assert actual_dict == {"children": []}


def test_keep_empty_adapter_local():
    @attrs.define
    class _Item(Item):
        children: CollectionType[Item]

    class TestAdapter(ItemAdapter):
        ADAPTER_CLASSES = [ZyteItemKeepEmptyAdapter] + list(ItemAdapter.ADAPTER_CLASSES)

    item = _Item([])
    adapter = TestAdapter(item)
    actual_dict = adapter.asdict()
    assert actual_dict == {"children": []}



import attrs

from zyte_common_items.base import Item


@attrs.define
class Brand(Item):
    """Brand.

    See :attr:`Product.brand <zyte_common_items.Product.brand>`.
    """

    #: Name as it appears on the source webpage (no post-processing).
    name: str



import datetime

from web_poet import RequestUrl

from zyte_common_items import (
    Address,
    AggregateRating,
    Amenity,
    BaseSalary,
    Breadcrumb,
    BusinessPlaceMetadata,
    Header,
    HiringOrganization,
    JobLocation,
    JobPostingMetadata,
    Link,
    Metadata,
    NamedLink,
    OpeningHoursItem,
    ParentPlace,
    ProbabilityMetadata,
    ProbabilityRequest,
    Reactions,
    RealEstateArea,
    Request,
    SocialMediaPostAuthor,
    SocialMediaPostMetadata,
    StarRating,
    Url,
)


def test_aggregated_rating_optional_fields():
    AggregateRating(bestRating=5.0)
    AggregateRating(ratingValue=2.5)
    AggregateRating(reviewCount=123)


def test_breadcrumb_optional_fields():
    Breadcrumb(name="foo")
    Breadcrumb(url="https://example.com")


def test_link_optional_fields():
    Link(text="foo")
    Link(url="https://example.com")


def test_metadata_default_values():
    metadata = Metadata()
    assert metadata.dateDownloaded is None
    assert metadata.get_date_downloaded_parsed() is None
    assert metadata.probability == 1.0


def test_metadata_get_date_downloaded():
    metadata = Metadata(dateDownloaded="2024-04-11T15:06:02Z")
    dt = metadata.get_date_downloaded_parsed()
    assert dt == datetime.datetime(
        year=2024,
        month=4,
        day=11,
        hour=15,
        minute=6,
        second=2,
        microsecond=0,
        tzinfo=datetime.timezone.utc,
    )


def test_named_link_optional_fields():
    NamedLink(name="foo")
    NamedLink(url="https://example.com")


def test_business_place_metadata_default_values():
    metadata = BusinessPlaceMetadata()
    assert metadata.dateDownloaded is None
    assert metadata.probability == 1.0
    assert metadata.searchText is None


def test_address():
    Address(addressRaw="2440 Hoonani Rd, Koloa, HI 96756, US")
    Address(streetAddress="2440 Hoonani Rd, Koloa, HI 96756")
    Address(addressCity="Koloa")
    Address(addressLocality="Poipu")
    Address(addressRegion="Kauaʻi County")
    Address(addressCountry="US")
    Address(postalCode="96756")
    Address(postalCodeAux="HI")
    Address(latitude=21.8768183)
    Address(longitude=-159.3620652)


def test_amenity():
    Amenity(name="WiFi", value=True)
    Amenity(name="WiFi", value=False)


def test_star_rating():
    StarRating(raw="4-star hotel")
    StarRating(ratingValue=4)


def test_parent_place():
    ParentPlace(name="Gateway Plaza", placeId="ChIJCVDfcEKuEmsR7_zYc0_GBtQ")


def test_opening_hours():
    OpeningHoursItem(dayOfWeek="Sunday", opens="9:00", closes="12:00")


def test_real_estate_area():
    RealEstateArea(value=123.45, unitCode="SQMT", raw="123.45 square meters")
    RealEstateArea(
        value=123.45, unitCode="SQMT", raw="123.45 square meters", areaType="LOT"
    )


def test_headers():
    Header(name="Connection", value="keep-alive")
    Header(name="Content-Type", value="text/html; charset=utf-8")


def test_probability_request():
    ProbabilityRequest(url="https://example.com")
    headers = [
        Header(name="Content-Type", value="application/x-www-form-urlencoded"),
        Header(name="Host", value="foo.example"),
    ]
    request = ProbabilityRequest(
        name="Post Test",
        url="https://example.com/test",
        method="POST",
        body="ZmllbGQxPXZhbHVlMSZmaWVsZDI9dmFsdWUy",
        headers=headers,
        metadata=ProbabilityMetadata(probability=0.5),
    )
    assert request.get_probability() == 0.5

    request = ProbabilityRequest(
        name="Get with RequestURL object",
        url=RequestUrl("https://example.com/test"),
    )

    assert request.url == "https://example.com/test"
    assert request.get_probability() is None


def test_request():
    Request(url="https://example.com")
    headers = [
        Header(name="Content-Type", value="application/x-www-form-urlencoded"),
        Header(name="Host", value="foo.example"),
    ]
    Request(
        name="Post Test",
        url="https://example.com/test",
        method="POST",
        body="ZmllbGQxPXZhbHVlMSZmaWVsZDI9dmFsdWUy",
        headers=headers,
    )

    request = Request(
        name="Get with RequestURL object",
        url=RequestUrl("https://example.com/test"),
    )

    assert request.url == "https://example.com/test"


def test_job_location():
    JobLocation(raw="New York, NY")


def test_base_salary():
    BaseSalary(
        raw="$53,000-$55,000 a year",
        valueMin="53000",
        valueMax="55000",
        rateType="yearly",
        currencyRaw="$",
        currency="USD",
    )


def test_hiring_organization():
    HiringOrganization(name="Example Inc.", nameRaw="Example Inc. USA", id="12345")


def test_job_posting_metadata():
    JobPostingMetadata(
        dateDownloaded="2020-01-01", probability=0.5, searchText="Software Engineer"
    )


def test_reactions():
    Reactions(reposts=1, likes=2, dislikes=3)


def test_social_media_post_author():
    SocialMediaPostAuthor(
        numberOfFollowers=5,
        numberOfFollowing=5,
        dateAccountCreated="2020-01-01",
        location="New York, NY",
        isVerified=True,
    )


def test_social_media_post_metadata():
    SocialMediaPostMetadata(
        dateDownloaded="2020-01-01", probability=0.5, searchText="Extract Summit"
    )


def test_url():
    Url(url="https://example.com")



from typing import List, Optional

import attrs
from web_poet import Returns

from zyte_common_items.components import Breadcrumb, Link
from zyte_common_items.fields import auto_field
from zyte_common_items.items import ProductFromList, ProductList, ProductListMetadata
from zyte_common_items.processors import breadcrumbs_processor

from .base import BasePage, Page
from .mixins import HasMetadata


class BaseProductListPage(
    BasePage, Returns[ProductList], HasMetadata[ProductListMetadata]
):
    """:class:`BasePage` subclass for :class:`ProductList`."""

    class Processors(BasePage.Processors):
        breadcrumbs = [breadcrumbs_processor]


class ProductListPage(Page, Returns[ProductList], HasMetadata[ProductListMetadata]):
    """:class:`Page` subclass for :class:`ProductList`."""

    class Processors(Page.Processors):
        breadcrumbs = [breadcrumbs_processor]


@attrs.define
class AutoProductListPage(BaseProductListPage):
    product_list: ProductList

    @auto_field
    def breadcrumbs(self) -> Optional[List[Breadcrumb]]:
        return self.product_list.breadcrumbs

    @auto_field
    def canonicalUrl(self) -> Optional[str]:
        return self.product_list.canonicalUrl

    @auto_field
    def categoryName(self) -> Optional[str]:
        return self.product_list.categoryName

    @auto_field
    def metadata(self) -> Optional[ProductListMetadata]:
        return self.product_list.metadata

    @auto_field
    def pageNumber(self) -> Optional[int]:
        return self.product_list.pageNumber

    @auto_field
    def paginationNext(self) -> Optional[Link]:
        return self.product_list.paginationNext

    @auto_field
    def products(self) -> Optional[List[ProductFromList]]:
        return self.product_list.products

    @auto_field
    def url(self) -> Optional[str]:
        return self.product_list.url



import attrs
import pytest
from parsel import Selector
from web_poet import field

from zyte_common_items import (
    ProductFromList,
    ProductFromListExtractor,
    ProductFromListSelectorExtractor,
    ProductVariant,
    ProductVariantExtractor,
    ProductVariantSelectorExtractor,
)

from .test_processors import gtin_expected, gtin_str


@pytest.mark.asyncio
async def test_product_from_list_extractor():
    @attrs.define
    class MyProductFromListExtractor(ProductFromListExtractor):
        selector: Selector

        @field
        def price(self):
            return self.selector.css("price")

        @field
        def regularPrice(self):
            return self.selector.css("oldPrice")

    selector = Selector("<data><price>10€</price><oldPrice>20€</oldPrice></data>")
    extracted = await MyProductFromListExtractor(selector).to_item()

    assert isinstance(extracted, ProductFromList)
    assert extracted.price == "10.00"
    assert extracted.regularPrice == "20.00"


@pytest.mark.asyncio
async def test_product_from_list_selector_extractor():
    class MyProductFromListSelectorExtractor(ProductFromListSelectorExtractor):
        @field
        def price(self):
            return self.css("price")

        @field
        def regularPrice(self):
            return self.css("oldPrice")

    selector = Selector("<data><price>10€</price><oldPrice>20€</oldPrice></data>")
    extracted = await MyProductFromListSelectorExtractor(selector).to_item()

    assert isinstance(extracted, ProductFromList)
    assert extracted.price == "10.00"
    assert extracted.regularPrice == "20.00"


@pytest.mark.asyncio
async def test_product_variant_extractor():
    @attrs.define
    class MyProductVariantExtractor(ProductVariantExtractor):
        selector: Selector

        @field
        def gtin(self):
            return self.selector.css("gtin")

        @field
        def price(self):
            return self.selector.css("price")

        @field
        def regularPrice(self):
            return self.selector.css("oldPrice")

    selector = Selector(
        f"<data>"
        f"<price>10€</price>"
        f"<oldPrice>20€</oldPrice>"
        f"<unusedField>foo</unusedField>"
        f"<gtin>{gtin_str}</gtin>"
        f"</data>"
    )
    extracted = await MyProductVariantExtractor(selector).to_item()

    assert isinstance(extracted, ProductVariant)
    assert extracted.gtin == gtin_expected
    assert extracted.price == "10.00"
    assert extracted.regularPrice == "20.00"


@pytest.mark.asyncio
async def test_product_variant_selector_extractor():
    class MyProductVariantSelectorExtractor(ProductVariantSelectorExtractor):
        @field
        def gtin(self):
            return self.css("gtin")

        @field
        def price(self):
            return self.css("price")

        @field
        def regularPrice(self):
            return self.css("oldPrice")

    selector = Selector(
        f"<data>"
        f"<price>10€</price>"
        f"<oldPrice>20€</oldPrice>"
        f"<unusedField>foo</unusedField>"
        f"<gtin>{gtin_str}</gtin>"
        f"</data>"
    )
    extracted = await MyProductVariantSelectorExtractor(selector).to_item()

    assert isinstance(extracted, ProductVariant)
    assert extracted.gtin == gtin_expected
    assert extracted.price == "10.00"
    assert extracted.regularPrice == "20.00"



from collections.abc import Iterable, Mapping
from functools import wraps
from numbers import Real
from typing import Any, Callable, List, Optional, Union

from clear_html import clean_node, cleaned_node_to_html, cleaned_node_to_text
from lxml.html import HtmlElement
from parsel import Selector, SelectorList
from price_parser import Price
from web_poet.mixins import ResponseShortcutsMixin
from zyte_parsers import Breadcrumb as zp_Breadcrumb
from zyte_parsers import Gtin as zp_Gtin
from zyte_parsers import (
    extract_brand_name,
    extract_breadcrumbs,
    extract_gtin,
    extract_price,
    extract_rating,
    extract_review_count,
)

from .components import (
    AggregateRating,
    BaseMetadata,
    Brand,
    Breadcrumb,
    Gtin,
    Image,
    ProbabilityRequest,
    Request,
)


def _get_base_url(page: Any) -> Optional[str]:
    if isinstance(page, ResponseShortcutsMixin):
        return page.base_url
    return getattr(page, "url", None)


def _handle_selectorlist(value: Any) -> Any:
    if not isinstance(value, SelectorList):
        return value
    if len(value) == 0:
        return None
    return value[0]


def _format_price(price: Price) -> Optional[str]:
    """Return the price amount as a string, with a minimum of 2 decimal
    places."""
    if price.amount is None:
        return None
    *_, exponent = price.amount.as_tuple()
    if not isinstance(exponent, int):
        return None  # NaN, Infinity, etc.
    if exponent <= -2:
        return str(price.amount)
    return f"{price.amount:.2f}"


def only_handle_nodes(
    f: Callable[[Union[Selector, HtmlElement], Any], Any]
) -> Callable[[Any, Any], Any]:
    """Decorator for processors that only runs a decorated processor if the
    input is of type :class:`Selector` or :class:`HtmlElement`."""

    @wraps(f)
    def wrapper(value: Any, page: Any) -> Any:
        value = _handle_selectorlist(value)
        if not isinstance(value, (Selector, HtmlElement)):
            return value
        result = f(value, page)
        return result

    return wrapper


def breadcrumbs_processor(value: Any, page: Any) -> Any:
    """Convert the data into a list of :class:`~zyte_common_items.Breadcrumb` objects if possible.

    Supported inputs are :class:`~parsel.selector.Selector`,
    :class:`~parsel.selector.SelectorList`, :class:`~lxml.html.HtmlElement` and
    an iterable of :class:`zyte_parsers.Breadcrumb` objects. Other inputs are
    returned as is.
    """

    def _from_zp_breadcrumb(value: zp_Breadcrumb) -> Breadcrumb:
        return Breadcrumb(name=value.name, url=value.url)

    value = _handle_selectorlist(value)

    if isinstance(value, (Selector, HtmlElement)):
        zp_breadcrumbs = extract_breadcrumbs(value, base_url=_get_base_url(page))
        return (
            [_from_zp_breadcrumb(b) for b in zp_breadcrumbs] if zp_breadcrumbs else None
        )

    if not isinstance(value, Iterable) or isinstance(value, str):
        return value

    results: List[Any] = []
    for item in value:
        if isinstance(item, zp_Breadcrumb):
            results.append(_from_zp_breadcrumb(item))
        else:
            results.append(item)
    return results


def brand_processor(value: Any, page: Any) -> Any:
    """Convert the data into a brand name if possible.

    If inputs are either :class:`~parsel.selector.Selector`,
    :class:`~parsel.selector.SelectorList` or :class:`~lxml.html.HtmlElement`, attempts
    to extract brand data from it.

    If value is a string, uses it to create a :class:`~zyte_common_items.Brand` instance.

    Other inputs are returned unchanged.
    """
    value = _handle_selectorlist(value)

    if isinstance(value, str):
        value = value.strip()
        return Brand(name=value) if value else None

    if isinstance(value, (Selector, SelectorList, HtmlElement)):
        if brand_name := extract_brand_name(value, search_depth=2):
            return Brand(name=brand_name)
        else:
            return None

    return value


def price_processor(value: Any, page: Any) -> Any:
    """Convert the data into a price string if possible.

    Uses the price-parser_ library.

    Supported inputs are :class:`~parsel.selector.Selector`,
    :class:`~parsel.selector.SelectorList`, :class:`~lxml.html.HtmlElement` and numeric values.

    Other inputs are returned as is.

    Puts the parsed Price object into ``page._parsed_price``.

    .. _price-parser: https://github.com/scrapinghub/price-parser
    """
    value = _handle_selectorlist(value)

    if isinstance(value, Real):
        return f"{value:.2f}"
    elif isinstance(value, (Selector, HtmlElement)):
        price = extract_price(value)
        page._parsed_price = price
        return _format_price(price)
    else:
        return value


def simple_price_processor(value: Any, page: Any) -> Any:
    """Convert the data into a price string if possible.

    Uses the price-parser_ library.

    Supported inputs are :class:`~parsel.selector.Selector`,
    :class:`~parsel.selector.SelectorList`, :class:`~lxml.html.HtmlElement` and numeric values.

    Other inputs are returned as is.

    .. _price-parser: https://github.com/scrapinghub/price-parser
    """
    value = _handle_selectorlist(value)

    if isinstance(value, Real):
        return f"{value:.2f}"
    elif isinstance(value, (Selector, HtmlElement)):
        price = extract_price(value)
        return _format_price(price)
    else:
        return value


@only_handle_nodes
def description_html_processor(value: Union[Selector, HtmlElement], page: Any) -> Any:
    """Convert the data into a cleaned up HTML if possible.

    Uses the clear-html_ library.

    Supported inputs are :class:`~parsel.selector.Selector`,
    :class:`~parsel.selector.SelectorList` and :class:`~lxml.html.HtmlElement`.
    Other inputs are returned as is.

    Puts the cleaned HtmlElement object into ``page._descriptionHtml_node``.

    .. _clear-html: https://github.com/zytedata/clear-html
    """
    if isinstance(value, Selector):
        value = value.root
    if value is None:
        return None
    if not isinstance(value, HtmlElement):
        raise ValueError(
            f"description_html_processor expects an HtmlElement node, got "
            f"{value.__class__}"
        )
    cleaned_node = clean_node(value, _get_base_url(page))
    page._descriptionHtml_node = cleaned_node
    return cleaned_node_to_html(cleaned_node)


def description_processor(value: Any, page: Any) -> Any:
    """Convert the data into a cleaned up text if possible.

    Uses the clear-html_ library.

    Supported inputs are :class:`~parsel.selector.Selector`,
    :class:`~parsel.selector.SelectorList` and :class:`~lxml.html.HtmlElement`.
    Other inputs are returned as is.

    Puts the cleaned HtmlElement object into ``page._description_node`` and the
    cleaned text into ``page._description_str``.

    .. _clear-html: https://github.com/zytedata/clear-html
    """
    value = _handle_selectorlist(value)
    if isinstance(value, str):
        page._description_str = value
        return value
    if isinstance(value, Selector):
        value = value.root
    if value is None:
        return None
    if not isinstance(value, HtmlElement):
        raise ValueError(
            f"description_processor expects an HtmlElement node, got "
            f"{value.__class__}"
        )
    cleaned_node = clean_node(value, _get_base_url(page))
    cleaned_text = cleaned_node_to_text(cleaned_node)
    page._description_node = cleaned_node
    page._description_str = cleaned_text
    return cleaned_text


def gtin_processor(
    value: Union[SelectorList, Selector, HtmlElement, str], page: Any
) -> Any:
    """Convert the data into a list of :class:`~zyte_common_items.Gtin` objects if possible.

    Supported inputs are :class:`str`, :class:`~parsel.selector.Selector`,
    :class:`~parsel.selector.SelectorList`, :class:`~lxml.html.HtmlElement`, an
    iterable of :class:`str` and an iterable of :class:`zyte_parsers.Gtin`
    objects.
    Other inputs are returned as is.
    """

    def _from_zp_gtin(zp_value: zp_Gtin) -> Gtin:
        return Gtin(type=zp_value.type, value=zp_value.value)

    results = []
    if isinstance(value, SelectorList):
        for sel in value:
            if result := extract_gtin(sel):
                results.append(_from_zp_gtin(result))
    elif isinstance(value, (Selector, HtmlElement, str)):
        if result := extract_gtin(value):
            results.append(_from_zp_gtin(result))
    elif isinstance(value, Iterable):
        for item in value:
            if isinstance(item, zp_Gtin):
                results.append(_from_zp_gtin(item))
            elif isinstance(item, str):
                results.append(_from_zp_gtin(extract_gtin(item)))
            else:
                results.append(item)
    else:
        return value
    return results or None


def rating_processor(value: Any, page: Any) -> Any:
    """Convert the data into an :class:`~zyte_common_items.AggregateRating`
    object if possible.

    Supported inputs are selector-like objects
    (:class:`~parsel.selector.Selector`,
    :class:`~parsel.selector.SelectorList`, or
    :class:`~lxml.html.HtmlElement`).

    The input can also be a dictionary with one or more of the
    :class:`~zyte_common_items.AggregateRating` fields as keys. The values for
    those keys can be either final values, to be assigned to the corresponding
    fields, or selector-like objects.

    If a returning dictionary is missing the ``bestRating`` field and
    ``ratingValue`` is a selector-like object, ``bestRating`` may be extracted.

    For example, for the following input HTML:

    .. code-block:: html

        <span class="rating">3.8 out of 5 stars</span>
        <a class="reviews">See all 7 reviews</a>

    You can use:

    .. code-block:: python

        @field
        def aggregateRating(self):
            return {
                "ratingValue": self.css(".rating"),
                "reviewCount": self.css(".reviews"),
            }

    To get:

    .. code-block:: python

        AggregateRating(
            bestRating=5.0,
            ratingValue=3.8,
            reviewCount=7,
        )
    """
    value = _handle_selectorlist(value)
    if isinstance(value, (Selector, HtmlElement)):
        zp_rating = extract_rating(value)
        result = AggregateRating(
            reviewCount=extract_review_count(value),
            bestRating=zp_rating.bestRating,
            ratingValue=zp_rating.ratingValue,
        )
        if result.reviewCount or result.bestRating or result.ratingValue:
            return result
        return None
    elif isinstance(value, dict):
        result = AggregateRating()

        review_count = _handle_selectorlist(value.get("reviewCount"))
        if isinstance(review_count, (Selector, HtmlElement)):
            result.reviewCount = extract_review_count(review_count)
        elif review_count is not None:
            result.reviewCount = int(review_count)

        rating_value = _handle_selectorlist(value.get("ratingValue"))
        if isinstance(rating_value, (Selector, HtmlElement)):
            zp_rating = extract_rating(rating_value)
            result.ratingValue = zp_rating.ratingValue
            result.bestRating = zp_rating.bestRating
        elif rating_value is not None:
            result.ratingValue = float(rating_value)

        if (best_rating := value.get("bestRating")) is not None:
            result.bestRating = float(best_rating)

        if result.reviewCount or result.bestRating or result.ratingValue:
            return result
        return None
    return value


def images_processor(value: Any, page: Any) -> Any:
    """Convert the data into a list of :class:`~zyte_common_items.Image`
    objects if possible.

    If the input is a string, it's used as a url for returning image object.

    If input is either an iterable of strings or mappings with "url" key, they are
    used to populate image objects.

    Other inputs are returned unchanged.
    """

    if isinstance(value, str):
        return [Image(url=value)]

    if isinstance(value, Iterable):
        results: List[Any] = []
        for item in value:
            if isinstance(item, Image):
                results.append(item)
            elif isinstance(item, Mapping):
                if url := item.get("url"):
                    results.append(Image(url=url))
            elif isinstance(item, str):
                results.append(Image(url=item))

        return results

    return value


def probability_request_list_processor(
    request_list: List[Request],
) -> List[ProbabilityRequest]:
    """Convert all objects in *request_list*, which are instances of
    :class:`Request` or a subclass, into instances of
    :class:`ProbabilityRequest`."""
    return [request.cast(ProbabilityRequest) for request in request_list]


def metadata_processor(metadata: Optional[BaseMetadata], page):
    """Processor for a metadata field that ensures that the output metadata
    object uses the metadata class declared by *page*."""
    if metadata is None or page.metadata_cls is None:
        return None
    return metadata.cast(page.metadata_cls)



from collections import deque
from typing import List, Optional
from warnings import warn

import attrs
from itemadapter import ItemAdapter as _ItemAdapter

from zyte_common_items import Article, ArticleList, JobPosting, Product, ProductList
from zyte_common_items.adapter import ZyteItemAdapter
from zyte_common_items.base import Item

warn(
    (
        "The zyte_common_items.ae module is a temporary module that helps "
        "speed up migrating from Zyte Automatic Extraction to Zyte API "
        "automatic extraction "
        "(https://docs.zyte.com/zyte-api/migration/zyte/autoextract.html). "
        "However, this module will eventually be removed. Please, update your "
        "code not to depend on this module anymore."
    ),
    DeprecationWarning,
    stacklevel=2,
)


class ItemAdapter(_ItemAdapter):
    ADAPTER_CLASSES = deque([ZyteItemAdapter])


def _convert_details_metadata(data):
    if "metadata" in data:
        _remove(data["metadata"], ["dateDownloaded"])
        if _is_not_none(data["metadata"], "probability"):
            data["probability"] = data["metadata"].pop("probability")
        _remove(data, ["metadata"])
    data.setdefault("probability", 1.0)


def _set_if_truthy(data, field, value):
    if value:
        data[field] = value


def _to_url_list(data, old_k, new_k):
    if _is_truthy_else_remove(data, old_k):
        entries = []
        for entry in data.pop(old_k):
            if _is_truthy(entry, "url"):
                entries.append(entry["url"])
        if entries:
            data[new_k] = entries


@attrs.define(kw_only=True)
class AEBreadcrumb(Item):
    name: Optional[str] = None
    link: Optional[str] = None


def _convert_authors(data):
    if _is_truthy_else_remove(data, "authors"):
        author = None
        author_list = []
        for _author in data.pop("authors") or []:
            if author is None and _is_truthy(_author, "nameRaw"):
                author = _author["nameRaw"]
            if _is_truthy(_author, "name"):
                author_list.append(_author["name"])
        _set_if_truthy(data, "author", author)
        _set_if_truthy(data, "authorList", author_list)


# https://docs.zyte.com/automatic-extraction/article.html#available-fields
@attrs.define(kw_only=True)
class AEArticle(Item):
    headline: Optional[str] = None
    datePublished: Optional[str] = None
    datePublishedRaw: Optional[str] = None
    dateModified: Optional[str] = None
    dateModifiedRaw: Optional[str] = None
    author: Optional[str] = None
    authorsList: List[str] = attrs.Factory(list)
    inLanguage: Optional[str] = None
    breadcrumbs: List[AEBreadcrumb] = attrs.Factory(list)
    mainImage: Optional[str] = None
    images: List[str] = attrs.Factory(list)
    description: Optional[str] = None
    articleBody: Optional[str] = None
    articleBodyHtml: Optional[str] = None
    articleBodyRaw: Optional[str] = None
    videoUrls: List[str] = attrs.Factory(list)
    audioUrls: List[str] = attrs.Factory(list)
    probability: float
    canonicalUrl: Optional[str] = None
    url: str

    @classmethod
    def from_item(cls, item: Item):
        assert isinstance(item, Article)
        data = ItemAdapter(item).asdict()
        _convert_details_metadata(data)
        _convert_authors(data)
        _convert_breadcrumbs(data)
        _convert_images(data)
        _to_url_list(data, "audios", "audioUrls")
        _to_url_list(data, "videos", "videoUrls")
        return super().from_dict(data)


# https://docs.zyte.com/automatic-extraction/article-list.html#individual-articles
@attrs.define(kw_only=True)
class AEArticleFromList(Item):
    headline: Optional[str] = None
    datePublished: Optional[str] = None
    datePublishedRaw: Optional[str] = None
    author: Optional[str] = None
    authorsList: List[str] = attrs.Factory(list)
    inLanguage: Optional[str] = None
    mainImage: Optional[str] = None
    images: List[str] = attrs.Factory(list)
    articleBody: Optional[str] = None
    url: Optional[str] = None
    probability: float


@attrs.define(kw_only=True)
class AEPaginationLink(Item):
    url: Optional[str] = None
    text: Optional[str] = None


def _convert_list_item_metadata(item):
    if _is_truthy_else_remove(item, "metadata"):
        if _is_not_none(item["metadata"], "probability"):
            item["probability"] = item["metadata"].pop("probability")
        del item["metadata"]
    item.setdefault("probability", 1.0)


# https://docs.zyte.com/automatic-extraction/article-list.html#available-fields
@attrs.define(kw_only=True)
class AEArticleList(Item):
    url: str
    articles: List[AEArticleFromList] = attrs.Factory(list)
    paginationNext: Optional[AEPaginationLink] = None
    paginationPrevious: Optional[AEPaginationLink] = None

    @classmethod
    def from_item(cls, item: Item):
        assert isinstance(item, ArticleList)
        data = ItemAdapter(item).asdict()
        if "articles" in data:
            for article in data["articles"]:
                _convert_authors(article)
                _convert_images(article)
                _convert_list_item_metadata(article)
        _remove(data, ["metadata"])
        return super().from_dict(data)


@attrs.define(kw_only=True)
class AELocation(Item):
    raw: Optional[str] = None


@attrs.define(kw_only=True)
class AEOrganization(Item):
    raw: Optional[str] = None


@attrs.define(kw_only=True)
class AESalary(Item):
    raw: Optional[str] = None
    value: Optional[float] = None
    currency: Optional[str] = None


def _rename(data, old_k, new_k):
    if _is_truthy_else_remove(data, old_k):
        data[new_k] = data.pop(old_k)


# https://docs.zyte.com/automatic-extraction/job-posting.html#available-fields
@attrs.define(kw_only=True)
class AEJobPosting(Item):
    title: Optional[str] = None
    datePosted: Optional[str] = None
    validThrough: Optional[str] = None
    description: Optional[str] = None
    descriptionHtml: Optional[str] = None
    employmentType: Optional[str] = None
    hiringOrganization: Optional[AEOrganization] = None
    baseSalary: Optional[AESalary] = None
    jobLocation: Optional[AELocation] = None
    probability: float
    url: str

    @classmethod
    def from_item(cls, item: Item):
        assert isinstance(item, JobPosting)
        data = ItemAdapter(item).asdict()
        _convert_details_metadata(data)
        _remove(data, ["datePublishedRaw"])
        _rename(data, "jobTitle", "title")
        _rename(data, "datePublished", "datePosted")
        if _is_truthy(data, "hiringOrganization"):
            _rename(data["hiringOrganization"], "name", "raw")
        if _is_truthy(data, "baseSalary"):
            _remove(data["baseSalary"], ["currency"])
            if _is_truthy_else_remove(data["baseSalary"], "valueMax"):
                data["baseSalary"]["value"] = float(data["baseSalary"].pop("valueMax"))
            _rename(data["baseSalary"], "currencyRaw", "currency")
        return super().from_dict(data)


@attrs.define(kw_only=True)
class AEAdditionalProperty(Item):
    name: str
    value: Optional[str] = None


@attrs.define(kw_only=True)
class AEGTIN(Item):
    type: str
    value: str


@attrs.define(kw_only=True)
class AEOffer(Item):
    price: Optional[str] = None
    currency: Optional[str] = None
    availability: Optional[str] = None
    regularPrice: Optional[str] = None


@attrs.define(kw_only=True)
class AERating(Item):
    ratingValue: Optional[float] = None
    bestRating: Optional[float] = None
    reviewCount: Optional[int] = None


def _remove(data, fields):
    for field in fields:
        if field in data:
            del data[field]


def _is_truthy(data, field):
    return field in data and data[field]


def _is_truthy_else_remove(data, field):
    if field not in data:
        return False
    if data[field]:
        return True
    else:
        del data[field]
        return False


def _is_not_none(data, field):
    return field in data and data[field] is not None


_OFFER_FIELD_MAP = {
    "price": "price",
    "regularPrice": "regularPrice",
    "availability": "availability",
    "currencyRaw": "currency",
}


def _convert_offer(data):
    offer = {}
    for old_k, new_k in _OFFER_FIELD_MAP.items():
        if _is_not_none(data, old_k):
            offer[new_k] = data.pop(old_k)
    if offer:
        data["offers"] = [offer]


def _convert_images(data):
    if _is_truthy_else_remove(data, "mainImage"):
        main_image = data.pop("mainImage")
        if _is_truthy(main_image, "url"):
            data["mainImage"] = main_image["url"]
    _to_url_list(data, "images", "images")


def _convert_breadcrumbs(data):
    if _is_truthy_else_remove(data, "breadcrumbs"):
        for entry in data["breadcrumbs"]:
            _rename(entry, "link", "url")


# https://docs.zyte.com/automatic-extraction/product.html#available-fields
@attrs.define(kw_only=True)
class AEProduct(Item):
    name: Optional[str] = None
    offers: List[AEOffer] = attrs.Factory(list)
    sku: Optional[str] = None
    mpn: Optional[str] = None
    gtin: List[AEGTIN] = attrs.Factory(list)
    brand: Optional[str] = None
    breadcrumbs: List[AEBreadcrumb] = attrs.Factory(list)
    mainImage: Optional[str] = None
    images: List[str] = attrs.Factory(list)
    description: Optional[str] = None
    descriptionHtml: Optional[str] = None
    aggregateRating: Optional[AERating] = None
    color: Optional[str] = None
    size: Optional[str] = None
    style: Optional[str] = None
    additionalProperty: List[AEAdditionalProperty] = attrs.Factory(list)
    hasVariants: List["AEProduct"] = attrs.Factory(list)
    probability: float
    canonicalUrl: Optional[str] = None
    url: str

    @classmethod
    def from_item(cls, item: Item):
        assert isinstance(item, Product)

        def convert(data):
            _remove(data, ["currency", "features"])
            _convert_details_metadata(data)
            _convert_offer(data)
            if _is_truthy_else_remove(data, "brand"):
                brand = data.pop("brand")
                if _is_truthy(brand, "name"):
                    data["brand"] = brand["name"]
            _convert_breadcrumbs(data)
            _convert_images(data)
            _rename(data, "additionalProperties", "additionalProperty")

        data = ItemAdapter(item).asdict()
        convert(data)
        if _is_truthy_else_remove(data, "variants"):
            for variant in data["variants"]:
                convert(variant)
                for field in ("probability", "url"):
                    if field not in variant:
                        variant[field] = data[field]
            data["hasVariants"] = data.pop("variants")
        return super().from_dict(data)


# https://docs.zyte.com/automatic-extraction/product-list.html#individual-products
@attrs.define(kw_only=True)
class AEProductFromList(Item):
    name: Optional[str] = None
    offers: List[AEOffer] = attrs.Factory(list)
    sku: Optional[str] = None
    brand: Optional[str] = None
    mainImage: Optional[str] = None
    images: List[str] = attrs.Factory(list)
    description: Optional[str] = None
    descriptionHtml: Optional[str] = None
    aggregateRating: Optional[AERating] = None
    probability: float
    url: Optional[str] = None


# https://docs.zyte.com/automatic-extraction/product-list.html#available-fields
@attrs.define(kw_only=True)
class AEProductList(Item):
    url: str
    products: List[AEProductFromList] = attrs.Factory(list)
    breadcrumbs: List[AEBreadcrumb] = attrs.Factory(list)
    paginationNext: Optional[AEPaginationLink] = None
    paginationPrevious: Optional[AEPaginationLink] = None

    @classmethod
    def from_item(cls, item: Item):
        assert isinstance(item, ProductList)
        data = ItemAdapter(item).asdict()
        if "products" in data:
            for product in data["products"]:
                _remove(product, ["currency"])
                _convert_offer(product)
                _convert_images(product)
                if _is_truthy_else_remove(product, "metadata"):
                    if _is_not_none(product["metadata"], "probability"):
                        product["probability"] = product["metadata"].pop("probability")
                    del product["metadata"]
                product.setdefault("probability", 1.0)
        _remove(data, ["metadata", "categoryName"])
        _convert_breadcrumbs(data)
        return super().from_dict(data)


_CONVERSION_MAP = {
    Article: AEArticle,
    ArticleList: AEArticleList,
    JobPosting: AEJobPosting,
    Product: AEProduct,
    ProductList: AEProductList,
}


def downgrade(item: Item) -> Item:
    if item.__class__ not in _CONVERSION_MAP:
        return item
    return _CONVERSION_MAP[item.__class__].from_item(item)  # type: ignore[attr-defined]



import warnings
from typing import Any, Callable, Dict, Optional, Tuple, Type, TypeVar
from warnings import warn
from weakref import WeakKeyDictionary

import attrs

# backwards compatibility imports
from ._dateutils import format_datetime  # noqa: F401
from .converters import MetadataCaster, url_to_str  # noqa: F401

# Caches the attribute names for attr.s classes.
_CLASS_ATTRS: WeakKeyDictionary = WeakKeyDictionary()


def split_in_unknown_and_known_fields(
    data: Optional[dict], item_cls: Type
) -> Tuple[Dict, Dict]:
    """
    Return a pair of dicts. The first one contains those elements not belonging to the
    attr class ``item_cls``. The second one contains the rest. That is, those
    attributes not belonging to ``item_cls`` class
    """
    data = data or {}
    if not attrs.has(item_cls):
        raise ValueError(f"The cls {item_cls} is not attrs class")
    if item_cls not in _CLASS_ATTRS:
        _CLASS_ATTRS[item_cls] = {field.name for field in attrs.fields(item_cls)}
    unknown, known = split_dict(data, lambda k: k in _CLASS_ATTRS[item_cls])
    return unknown, known


def split_dict(dict: Dict, key_pred: Callable[[Any], Any]) -> Tuple[Dict, Dict]:
    """Splits the dictionary in two.

    The first dict contains the records for which the key predicate is False
    and the second dict contains the rest.

    >>> split_dict({}, lambda k: False)
    ({}, {})
    >>> split_dict(dict(a=1, b=2, c=3), lambda k: k != 'a')
    ({'a': 1}, {'b': 2, 'c': 3})
    """  # noqa
    yes, no = {}, {}
    for k, v in dict.items():
        if key_pred(k):
            yes[k] = v
        else:
            no[k] = v
    return (no, yes)


NewClassT = TypeVar("NewClassT", bound=attrs.AttrsInstance)


def convert_to_class(value: Any, new_cls: Type[NewClassT]) -> NewClassT:
    """Convert *value* into *type* keeping all shared attributes, and
    triggering a run-time warning if any attribute is removed."""
    if type(value) is new_cls:
        return value
    input_attributes = {attribute.name for attribute in attrs.fields(value.__class__)}
    output_attributes = {attribute.name for attribute in attrs.fields(new_cls)}
    shared_attributes = input_attributes & output_attributes
    new_value = new_cls(
        **{attribute: getattr(value, attribute) for attribute in shared_attributes}
    )
    removed_nonempty_attributes = {
        attribute
        for attribute in (input_attributes - output_attributes)
        if getattr(value, attribute)
        != attrs.fields_dict(value.__class__)[attribute].default
    }
    if removed_nonempty_attributes:
        warn(
            (
                f"Conversion of {value} into {new_cls} is dropping the non-default "
                f"values of the following attributes: "
                f"{removed_nonempty_attributes}."
            ),
            RuntimeWarning,
        )
    return new_value


def metadata_processor(metadata, page):
    from zyte_common_items.processors import metadata_processor

    warnings.warn(
        "zyte_common_items.util.metadata_processor is moved to"
        "zyte_common_items.processors.metadata_processor",
        DeprecationWarning,
        stacklevel=2,
    )
    return metadata_processor(metadata, page)



from typing import Optional

import attrs

from zyte_common_items.base import Item


@attrs.define(kw_only=True)
class RealEstateArea(Item):
    """Area of a place, with type, units, value and raw value."""

    #: Area
    value: float

    #: Unit of the value field, one of: SQMT (square meters), SQFT (square
    #: feet), ACRE (acres).
    unitCode: str

    #: Type of area, one of: LOT, FLOOR
    areaType: Optional[str] = None

    #: Area in the raw format, as it appears on the website.
    raw: str



from typing import List, Optional

import attrs
from web_poet import Returns

from zyte_common_items.components import BaseSalary, HiringOrganization, JobLocation
from zyte_common_items.fields import auto_field
from zyte_common_items.items import JobPosting, JobPostingMetadata
from zyte_common_items.processors import (
    description_html_processor,
    description_processor,
)

from .base import BasePage, Page
from .mixins import DescriptionMixin, HasMetadata


class BaseJobPostingPage(
    BasePage, DescriptionMixin, Returns[JobPosting], HasMetadata[JobPostingMetadata]
):
    """:class:`BasePage` subclass for :class:`JobPosting`."""

    class Processors(BasePage.Processors):
        description = [description_processor]
        descriptionHtml = [description_html_processor]


class JobPostingPage(
    Page, DescriptionMixin, Returns[JobPosting], HasMetadata[JobPostingMetadata]
):
    """:class:`Page` subclass for :class:`JobPosting`."""

    class Processors(Page.Processors):
        description = [description_processor]
        descriptionHtml = [description_html_processor]


@attrs.define
class AutoJobPostingPage(BaseJobPostingPage):
    job_posting: JobPosting

    @auto_field
    def url(self) -> Optional[str]:
        return self.job_posting.url

    @auto_field
    def jobPostingId(self) -> Optional[str]:
        return self.job_posting.jobPostingId

    @auto_field
    def datePublished(self) -> Optional[str]:
        return self.job_posting.datePublished

    @auto_field
    def datePublishedRaw(self) -> Optional[str]:
        return self.job_posting.datePublishedRaw

    @auto_field
    def dateModified(self) -> Optional[str]:
        return self.job_posting.dateModified

    @auto_field
    def dateModifiedRaw(self) -> Optional[str]:
        return self.job_posting.dateModifiedRaw

    @auto_field
    def validThrough(self) -> Optional[str]:
        return self.job_posting.validThrough

    @auto_field
    def validThroughRaw(self) -> Optional[str]:
        return self.job_posting.validThroughRaw

    @auto_field
    def jobTitle(self) -> Optional[str]:
        return self.job_posting.jobTitle

    @auto_field
    def headline(self) -> Optional[str]:
        return self.job_posting.headline

    @auto_field
    def jobLocation(self) -> Optional[JobLocation]:
        return self.job_posting.jobLocation

    @auto_field
    def description(self) -> Optional[str]:
        return self.job_posting.description

    @auto_field
    def descriptionHtml(self) -> Optional[str]:
        return self.job_posting.descriptionHtml

    @auto_field
    def employmentType(self) -> Optional[str]:
        return self.job_posting.employmentType

    @auto_field
    def baseSalary(self) -> Optional[BaseSalary]:
        return self.job_posting.baseSalary

    @auto_field
    def requirements(self) -> Optional[List[str]]:
        return self.job_posting.requirements

    @auto_field
    def hiringOrganization(self) -> Optional[HiringOrganization]:
        return self.job_posting.hiringOrganization

    @auto_field
    def jobStartDate(self) -> Optional[str]:
        return self.job_posting.jobStartDate

    @auto_field
    def jobStartDateRaw(self) -> Optional[str]:
        return self.job_posting.jobStartDateRaw

    @auto_field
    def remoteStatus(self) -> Optional[str]:
        return self.job_posting.remoteStatus

    @auto_field
    def metadata(self) -> Optional[JobPostingMetadata]:
        return self.job_posting.metadata



import attrs
from web_poet import ItemPage, RequestUrl, WebPage, field
from web_poet.pages import ItemT

from .._dateutils import utcnow_formatted
from ..processors import metadata_processor
from .mixins import HasMetadata, MetadataT


class _BasePage(ItemPage[ItemT], HasMetadata[MetadataT]):
    class Processors:
        metadata = [metadata_processor]

    @field
    def metadata(self) -> MetadataT:
        if self.metadata_cls is None:
            raise ValueError(f"{type(self)} doesn'have a metadata class configured.")
        value = self.metadata_cls()
        attributes = dir(value)
        if "dateDownloaded" in attributes:
            value.dateDownloaded = utcnow_formatted()  # type: ignore
        if "probability" in attributes:
            value.probability = 1.0  # type: ignore
        return value

    def no_item_found(self) -> ItemT:
        """Return an item with the current url and probability=0,
        indicating that the passed URL doesn't contain the expected item.

        Use it in your .validate_input implementation.
        """
        if self.metadata_cls is None:
            raise ValueError(f"{type(self)} doesn'have a metadata class configured.")
        metadata = self.metadata_cls()
        metadata_attributes = dir(metadata)
        if "dateDownloaded" in metadata_attributes:
            metadata.dateDownloaded = utcnow_formatted()  # type: ignore
        if "probability" in metadata_attributes:
            metadata.probability = 0.0  # type: ignore
        return self.item_cls(  # type: ignore
            url=self.url,  # type: ignore[attr-defined]
            metadata=metadata,
        )


@attrs.define
class BasePage(_BasePage):
    """Base class for page object classes that has
    :class:`~web_poet.page_inputs.http.RequestUrl` as a dependency."""

    class Processors(_BasePage.Processors):
        pass

    request_url: RequestUrl

    @field
    def url(self) -> str:
        return str(self.request_url)


@attrs.define
class Page(_BasePage, WebPage):
    """Base class for page object classes that has
    :class:`~web_poet.page_inputs.http.HttpResponse` as a dependency."""

    class Processors(_BasePage.Processors):
        pass

    @field
    def url(self) -> str:
        return str(self.response.url)
